import signal
from typing import Callable, Optional

from easy_pysy.utils.common import require
from easy_pysy.core import logging

exiting = False
_sigint_callback: Optional[Callable[[int], None]] = None


def sigint_callback(func):
    global _sigint_callback
    require(_sigint_callback is None, "There's already a sigint callback")
    _sigint_callback = func
    return func


def set_sigint_callback(callback: Callable[[int], None]):
    global _sigint_callback
    _sigint_callback = callback


def _sigint_handler(signum, frame):
    logging.info(f'Received sigint: {signum}')

    global exiting
    if exiting:
        logging.warning('Double sigint received, forcing exit')
        exit(1)

    exiting = True
    if _sigint_callback is not None:
        _sigint_callback(signum)


signal.signal(signal.SIGINT, _sigint_handler)



import logging

from typing import Optional
from uuid import uuid4


logger = logging.getLogger(__name__)


def uuid() -> str:
    return str(uuid4())


def require(condition: bool, message: Optional[str] = None, exception: Optional[BaseException] = None):
    if not condition:
        exception = exception or RequirementError(message)
        raise exception


class RequirementError(Exception):
    pass


class IntSequence:
    _last_id = 0

    def create_new_id(self) -> int:
        self._last_id += 1
        return self._last_id



from functools import partial
from typing import Callable, TypeVar, Any


T = TypeVar('T')


def bind(something: T, function: Callable[[T, ], Any]):
    self_function = partial(function, something)
    setattr(something, function.__name__, self_function)


def bind_all(something: T, functions: list[Callable[[T, ], Any]]):
    for function in functions:
        bind(something, function)



import time
from threading import Thread, Event, Timer
from typing import Callable, Optional

from easy_pysy.utils.common import require
from easy_pysy.core.configuration import config
from easy_pysy.core.logging import logger
from easy_pysy.core.lifecycle import AppStopping
from easy_pysy.core.event import on

stop_timeout = config('ez.thread.stop_timeout', config_type=int, default=1)


class EzThread(Thread):  # TODO: test
    def __init__(self, target, args=(), kwargs=None, name=None, daemon=False):
        if kwargs is None:
            kwargs = {}
        self.stop_event = Event()
        super().__init__(None, target, name, args, kwargs, daemon=daemon)
        self.target = target

    def start(self):
        require(not self.is_alive(), 'Interval already started')
        self.stop_event.clear()
        threads.append(self)
        super().start()

    def stop(self, timeout=10):
        require(self.is_alive(), 'Thread already stopped')
        self.stop_event.set()
        self.join(timeout)

    def join(self, timeout=None):
        super().join(timeout)
        if self.is_alive():
            assert timeout is not None
            # timeout can't be None here because if it was, then join()
            # wouldn't return until the thread was dead
            raise ZombieThread(f"Thread failed to die within {timeout} seconds")
        else:
            threads.remove(self)


class Interval(Timer):
    def __init__(self, interval_ms, function, on_error, args=(), kwargs=None):
        super().__init__(
            interval_ms / 1000.0,
            function,
            args=args,
            kwargs=kwargs or {},
        )
        self.on_error = on_error

    def run(self):
        next_time = time.time() + self.interval
        wait_time = next_time - time.time()
        while not self.finished.wait(wait_time):
            try:
                next_time += self.interval
                self.function(*self.args, **self.kwargs)
                wait_time = next_time - time.time()
            except BaseException as exc:
                self.on_error(exc)
                wait_time = next_time - time.time()


class ZombieThread(Exception):
    pass


threads: list[EzThread] = []


def get_thread(target: Callable) -> Optional[EzThread]:
    for t in threads:
        if t.target == target:
            return t
    return None


@on(AppStopping)
def on_stop(event: AppStopping):
    logger.debug('Stopping threads')
    while threads:
        running_threads = [thread for thread in threads if thread.is_alive()]
        # Optimization: instead of calling stop() which implicitly calls
        # join(), set all the stopping events simultaneously, *then* join
        # threads with a reasonable timeout
        for thread in running_threads:
            thread.stop_event.set()
        for thread in running_threads:
            logger.debug(f'Stopping thread: {thread}')
            thread.join(stop_timeout)
    logger.debug('Threads stopped')



import os
from datetime import datetime, date
from typing import Type, cast, Union

import pendulum
from dotenv import load_dotenv, find_dotenv

# TODO: dict and list
SupportedTypes = Union[str, int, float, bool, datetime, date]

dotenv_path = os.getenv('DOTENV_PATH') or find_dotenv(usecwd=True)
loaded = load_dotenv(dotenv_path)


def config(key: str, config_type: Type[SupportedTypes] = str, default=None, raise_if_not_found=False) -> SupportedTypes:
    raw = os.getenv(key)

    if raw is None and raise_if_not_found:
        raise ConfigNotFoundError(f'Config {key} not found')
    elif raw is None:
        return default

    if config_type == str:
        return raw
    elif config_type == int:
        return int(raw)
    elif config_type == float:
        return float(raw)
    elif config_type == bool:
        return raw.lower() == 'true' or raw == '1'
    elif config_type == datetime:
        return cast(datetime, pendulum.parse(raw))
    elif config_type == date:
        return cast(datetime, pendulum.parse(raw)).date()
    else:
        raise TypeNotSupportedError(f'Type {config_type} is not supported for configuration')


class ConfigNotFoundError(Exception):
    pass


class TypeNotSupportedError(Exception):
    pass



import csv
from typing import Iterable, Callable, Any


def read_md_table(lines: Iterable[str], types: dict[str, Callable[[str], Any]] = None) -> Iterable[dict]:
    # Cleanup: Remove leading and trailing '|'
    lines = map(lambda line: line.lstrip('|').rstrip('|\n'), lines)

    # We use the builtin csv reader
    reader = csv.DictReader(lines, delimiter='|', quoting=csv.QUOTE_NONE)

    # Remove first header line "-----"
    next(reader)

    # Cleanup dicts (strip keys and values)
    items = map(lambda obj: _cleanup(obj), reader)

    # Post Processing
    if types:
        items = map(lambda obj: _post_process(obj, types), items)

    return items


def _post_process(obj: dict, types: dict[str, Callable[[str], Any]]):
    for key, t in types.items():
        obj[key] = t(obj[key]) if obj[key] != '' else None
    return obj


def _cleanup(obj: dict):
    return {key.strip(): value.strip() for key, value in obj.items()}



from dataclasses import dataclass
from typing import TypeVar, Type, Callable, Generic, Any

import easy_pysy.utils.decorators
from easy_pysy.core import logging

T = TypeVar('T')
ProviderFactory = Callable[[], T]


@dataclass
class Provider(Generic[T]):
    type: Type[T]
    factory: ProviderFactory
    singleton: bool


providers: dict[Type, list[Provider]] = {}
singletons: dict[Type, Any] = {}


def provide(type: Type[T], singleton: bool = False):
    # TODO: env and variant
    def decorator(func):
        new_provider = Provider(type, func, singleton)
        if type not in providers:
            providers[type] = [new_provider]
        else:
            providers[type].append(new_provider)
            logging.warning(f'Multiple providers defined for {type}: {providers[type]}')

        return func
    return decorator


def get(type: Type[T]) -> T:  # TODO: other name
    easy_pysy.utils.decorators.require(type in providers and providers[type], f"No provider found for {type}")
    provider = providers[type][-1]

    if not provider.singleton:
        return provider.factory()
    elif type not in singletons:
        instance = provider.factory()
        singletons[type] = instance
        return instance
    else:
        return singletons[type]



from easy_pysy.utils.functional.dictionary import EzDict, K, V
from easy_pysy.utils.functional.sequence import EzList, T
from easy_pysy.utils.functional.function import bind, bind_all

from typing import Iterable, Union


def magic(value: Union[list[T], dict[K, V]]) -> Union[EzList[T], EzDict[K, V]]:  # TODO: iterable/sequence
    if isinstance(value, dict):
        return EzDict(value)
    elif isinstance(value, Iterable):
        return EzList(value)
    else:
        raise NotImplementedError(f'Unsupported magic: {type(value)}')



import typer
from easy_pysy.core.lifecycle import start, context, AppState

main_typer = typer.Typer()
# main_typer.add_typer(run_typer, name="run")

command = main_typer.command


def run(auto_start=True):
    if auto_start and context.state == AppState.STOPPED:
        start()
    main_typer()



from typing import Any


def qual_name(obj: Any):
    object_type = type(obj)
    module = object_type.__module__
    if module == 'builtins':
        return object_type.__qualname__
    return module + '.' + object_type.__qualname__



def tri_wave(min_value: int, max_value: int, step: int = 1):
    while True:
        yield from range(min_value, max_value, step)
        yield from range(max_value, min_value, -step)


def float_range(start: float, stop: float, step: float = 1.0, decimals: int = 2):
    for i in range(int(start / step), int(stop / step)):
        yield round(i * step, ndigits=decimals)
